import os
import shutil
from typing import Literal

from .print import debug


def ensure_dir(dir: str, if_not_empty: Literal['keep', 'clean', 'error'] = 'keep'):
    """检查用于保存日志、Checkpoint 等的目录如果 dir 不存在，则创建

    Parameters
    ----------
    - `dir` : `str`
        - 目录
    - `if_not_empty` : `Literal['keep', 'clean', 'error']`, optional
        - 如果目录不为空，则执行的操作
        - `"keep"` | `"clean"` | `"error"`
    """
    if not os.path.exists(dir):
        os.makedirs(dir)
    elif any(os.scandir(dir)):
        if if_not_empty is None or if_not_empty == 'keep':
            debug(f'Exists: {dir}!')
        elif if_not_empty == 'clean':
            debug(f'Clean up: {dir}!')
            shutil.rmtree(dir)
            os.makedirs(dir)
        elif if_not_empty == 'error':
            raise Exception(f'目录 {dir} 不为空!')
        else:
            raise ValueError(f'check_dir 参数错误: if_not_empty = {if_not_empty}')



import torch
from torch import nn


class DepthwiseSeperableConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        bias=False,
    ):
        super(DepthwiseSeperableConv, self).__init__()
        self.depthwise_conv = nn.Conv1d(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            in_channels,
            bias,
        )
        self.pointwise_conv = nn.Conv1d(in_channels, out_channels, 1, 1, 0, 1, 1, bias)

    def forward(self, x):
        x = self.depthwise_conv(x)
        x = self.pointwise_conv(x)
        return x


# ========================= 2D ========================= #


class DepthwiseSeperableConv2D(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        bias=False,
    ):
        super(DepthwiseSeperableConv2D, self).__init__()
        self.depthwise_conv = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            in_channels,
            bias,
        )
        self.pointwise_conv = nn.Conv2d(in_channels, out_channels, 1, 1, 0, 1, 1, bias)

    def forward(self, x):
        x = self.depthwise_conv(x)
        x = self.pointwise_conv(x)
        return x



import torch
from torch.utils.data import Dataset, DataLoader, random_split


def train_val_split(
    data_set: Dataset, val_proportion: float, n_batch: int
) -> (DataLoader, DataLoader):
    """Split DataSet into TrainDataLoader and ValDataLoader"""
    n_data = len(data_set)
    n_val = int(n_data * val_proportion)
    n_train = n_data - n_val
    train_set, val_set = random_split(data_set, [n_train, n_val])
    train_data_loader = DataLoader(
        train_set, batch_size=n_batch, drop_last=True, num_workers=1, pin_memory=True
    )
    val_data_loader = DataLoader(
        val_set, batch_size=n_batch, drop_last=True, num_workers=1, pin_memory=True
    )
    return train_data_loader, val_data_loader



from torch import nn
from torch.nn import functional as F


def build_norm(type=None, channels=None):
    assert type is None or type in ['batch', 'instance', 'pixel']
    if type is None:
        return nn.Identity()
    elif type == 'batch':
        return nn.BatchNorm1d(channels)
    elif type == 'instance':
        return nn.InstanceNorm1d(channels)
    elif type == 'pixel':
        return PixelNorm()


class PixelNorm(nn.Module):
    """Pixel Normalization from

    'PROGRESSIVE GROWING OF GANS FOR IMPROVED QUALITY, STABILITY, AND VARIATION

    沿着通道维度对 Pixel 向量做标准化。如输入为 `[N, 50, 1024]`
    则该方法对 N 个样本中，所有 1024 个点在 50 个通道上组成的 50-dim 向量做归一化
    """

    def __init__(self):
        super(PixelNorm, self).__init__()

    def forward(self, x, epsilon=1e-8):
        # 沿着 Channel 维，对每个 pixel 做一次 norm
        assert len(x.shape) == 3
        return x * (((x**2).mean(dim=1, keepdim=True) + epsilon).rsqrt())


class ConditionalBatchNorm1d(nn.BatchNorm1d):
    """Conditional Batch Normalization"""

    def __init__(
        self,
        num_features,
        eps=1e-05,
        momentum=0.1,
        affine=False,
        track_running_stats=True,
    ):
        super(ConditionalBatchNorm1d, self).__init__(
            num_features, eps, momentum, affine, track_running_stats
        )

    def forward(self, input, weight, bias, **kwargs):
        self._check_input_dim(input)

        exponential_average_factor = 0.0

        if self.training and self.track_running_stats:
            self.num_batches_tracked += 1
            if self.momentum is None:  # use cumulative moving average
                exponential_average_factor = 1.0 / self.num_batches_tracked.item()
            else:  # use exponential moving average
                exponential_average_factor = self.momentum

        output = F.batch_norm(
            input,
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            self.training or not self.track_running_stats,
            exponential_average_factor,
            self.eps,
        )
        if weight.dim() == 1:
            weight = weight.unsqueeze(0)
        if bias.dim() == 1:
            bias = bias.unsqueeze(0)
        size = output.size()
        weight = weight.unsqueeze(-1).expand(size)
        bias = bias.unsqueeze(-1).expand(size)
        return weight * output + bias


class CategoricalConditionalBatchNorm1d(ConditionalBatchNorm1d):
    def __init__(
        self,
        num_classes,
        num_features,
        eps=1e-5,
        momentum=0.1,
        affine=False,
        track_running_stats=True,
    ):
        super(CategoricalConditionalBatchNorm1d, self).__init__(
            num_features, eps, momentum, affine, track_running_stats
        )
        self.weights = nn.Embedding(num_classes, num_features)
        self.biases = nn.Embedding(num_classes, num_features)

        self._initialize()

    def _initialize(self):
        nn.init.ones_(self.weights.weight.data)
        nn.init.zeros_(self.biases.weight.data)

    def forward(self, input, c, **kwargs):
        weight = self.weights(c)
        bias = self.biases(c)
        return super(CategoricalConditionalBatchNorm1d, self).forward(input, weight, bias)



from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Set, TypedDict, Union
from xml.etree import ElementTree as ET

from zyplib.annotation.event import Event


@dataclass
class ScoredEventType:
    type: int
    name: str

    def __hash__(self) -> int:
        return self.type

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, ScoredEventType):
            return self.type == __value.type
        elif isinstance(__value, int):
            return self.type == __value
        elif isinstance(__value, str):
            return self.name == __value
        else:
            return False


@dataclass
class ScoredEvent(ScoredEventType):
    type: int
    name: str
    time: float
    duration: float
    manully_scored: int
    param1: Optional[int] = None
    param2: Optional[int] = None
    param3: Optional[int] = None

    def same_type(self, __value: Union[int, str, ScoredEventType]) -> bool:
        return super().__eq__(__value)

    def to_event(self) -> Event:
        return Event(name=self.name, onset=self.time, duration=self.duration)


class PSGScoredEvents:
    """CMPPSGSCOREDATA XML 文件的读取"""

    def __init__(self) -> None:
        self.base_name: str = None  # 如 'cyq202107220206' 包含了姓名和基本信息
        self.created_on: datetime = None
        self.scored_events: List[ScoredEvent] = []
        self.sleep_stages: List[int] = []

    def push_event_(self, onset, duration, name, type=-1, manually_scored=-1):
        """添加事件"""
        scored_event = ScoredEvent(
            type=type,
            name=name,
            time=onset,
            duration=duration,
            manully_scored=manually_scored,
        )
        self.scored_events.append(scored_event)

    @property
    def unique_events(self) -> Set[ScoredEventType]:
        """获取所有的唯一的事件"""
        return set(
            [
                ScoredEventType(type=score_event.type, name=score_event.name)
                for score_event in self.scored_events
            ]
        )

    def filter(self, *event_types: Union[int, str, ScoredEventType]) -> 'PSGScoredEvents':
        """根据事件类型过滤 scored_events, 并创建一个新的 PSGScoredData

        根据输入的 event_type 来过滤 scored_events, 并创建一个新的 PSGScoredData

        Parameters
        ----------
            - `event_types`: 事件类型, 可以是 int, str, ScoredEventType

        Returns
        ---------
            - `PSGScoredData`: 过滤后的 PSGScoredData
        """
        event_types = set(event_types)
        new_scored_events = []
        for score_event in self.scored_events:
            for event_type in event_types:
                if score_event.same_type(event_type):
                    # print('Same type', score_event)
                    new_scored_events.append(score_event)
        psg_scored_data = PSGScoredEvents()
        psg_scored_data.created_on = self.created_on
        psg_scored_data.scored_events = new_scored_events
        psg_scored_data.sleep_stages = self.sleep_stages
        return psg_scored_data

    def filter_(self, *event_types: Union[int, str, ScoredEventType]):
        """据事件类型过滤 scored_events, 同 `filter`, 但是直接更改 self"""
        psg_scored_data = self.filter(*event_types)
        self.created_on = psg_scored_data.created_on
        self.scored_events = psg_scored_data.scored_events
        self.sleep_stages = psg_scored_data.sleep_stages

    @staticmethod
    def from_xml_file(xml_fpath: str):
        fname = Path(xml_fpath).name
        base_name = fname.split('.')[0]

        # 1. 读取 xml 文件
        tree = ET.parse(xml_fpath)
        # 2. 读取 /SCOREDEVENTS/SCOREDEVENT
        root = tree.getroot()

        score_events = []
        SCOREDEVENTS = root.find('SCOREDEVENTS')
        for scoredevent in SCOREDEVENTS.iter('SCOREDEVENT'):
            score_event = ScoredEvent(
                type=int(scoredevent.find('TYPE').text),
                name=scoredevent.find('NAME').text,
                time=float(scoredevent.find('TIME').text),
                duration=float(scoredevent.find('DURATION').text),
                manully_scored=int(scoredevent.find('MANUALLYSCORED').text),
                param1=int(scoredevent.find('PARAM1').text),
                param2=int(scoredevent.find('PARAM2').text),
                param3=int(scoredevent.find('PARAM3').text),
            )
            score_events.append(score_event)
        # 3. 读取 /SLEEPSTAGES/SLEEPSTAGE
        sleep_stages = []
        for sleepstage in root.findall('SLEEPSTAGES/SLEEPSTAGE'):
            sleep_stages.append(int(sleepstage.text))
        # 4. 返回
        psg_scored_data = PSGScoredEvents()
        created_on_str = root.find('CREATEDON')  # 例如: 2021/7/22 19:47:00
        created_on: datetime = datetime.strptime(created_on_str.text, '%Y/%m/%d %H:%M:%S')
        psg_scored_data.created_on = created_on
        psg_scored_data.scored_events = score_events
        psg_scored_data.sleep_stages = sleep_stages
        psg_scored_data.base_name = base_name
        return psg_scored_data


def read_scoreddata_xml(xml_fpath: str) -> List[Event]:
    psg_events = PSGScoredEvents.from_xml_file(xml_fpath)
    return [event.to_event() for event in psg_events.scored_events]



from functools import singledispatch

from dataclasses import dataclass
from typing import List, Dict, Tuple, Union

import numpy as np


@dataclass
class Event:
    name: Union[str, int]
    onset: float
    duration: float


def filter_event(events: List[Event], name: Union[str, int]) -> List[Event]:
    return [event for event in events if event.name == name]


def split_event(events: List[Event]) -> Dict[str, List[Event]]:
    """根据 events 中的事件名称将 events 划分为多个事件列表"""
    events_dict: Dict[str, List[Event]] = {}
    for event in events:
        if event.name not in events_dict:
            events_dict[event.name] = []
        events_dict[event.name].append(event)
    return events_dict


def merge_events(events: list[Event], distance: float) -> list[Event]:
    """将距离小于 distance 的事件合并为一个事件

    本方法无视事件名称, 仅根据 onset 和 duration 进行合并

    Parameters
    ----------
        - `events` : list[Event]
            - _description_
        - `distance` : float
            - _description_

    Returns
    ----------
        - `list[Event]`
            - _description_
    """
    # 1. 将事件按 onset 排序
    events = sorted(events, key=lambda x: x.onset)
    # 2. 合并事件
    merged_events: List[Event] = []
    for event in events:
        if len(merged_events) == 0:
            merged_events.append(event)
        else:
            last_event = merged_events[-1]
            last_event_end = last_event.onset + last_event.duration
            if event.onset - last_event_end > distance:
                merged_events.append(event)
            else:
                last_event.duration = event.onset + event.duration - last_event.onset
    return merged_events


def events_from_array(array: Union[List[List[float]], np.ndarray], name) -> List[Event]:
    """从数组中提取事件列表

    Examples:
    ----------
    >>> events_from_array([[0, 1], [1, 2]], 'spindle')
    >>> events_from_array(np.array([[0, 1], [1, 2]]), 'spindle')

    Parameters
    ----------
        - `array` : Union[List[List[float]], np.ndarray]
            - 原始的事件列表

    Returns
    ----------
        - `List[Event]`
            - 事件对象列表
    """
    events = []
    for l in array:
        events.append(Event(name, l[0], l[1]))
    return events


def events_to_dict(events: List[Event]) -> Dict:
    dic = {
        'Onset': np.array([event.onset for event in events]),
        'Duration': np.array([event.duration for event in events]),
        'Name': np.array([event.name for event in events]),
    }
    return dic


def events_to_list(events: List[Event]) -> List[List[float]]:
    """将事件列表转换为数组

    Parameters
    ----------
        - `events` : List[Event]
            - 事件列表, Evnet 为 zyplib.annotation.type.Event 类型; 单位为秒

    Returns
    ----------
        - `list` : List[[float, float]]
            - 事件列表, [[onset, duration], ...]
    """
    return [[event.onset, event.duration] for event in events]


def events_to_npy(events: List[Event]) -> np.ndarray:
    """将事件列表转换为数组

    Parameters
    ----------
        - `events` : List[Event]
            - 事件列表, Evnet 为 zyplib.annotation.type.Event 类型; 单位为秒

    Returns
    ----------
        - `np.ndarray` : np.ndarray
            - 事件列表, [[onset, duration], ...]
    """
    return np.array(events_to_list(events))



import numpy as np
from scipy import signal, stats


def peak_num(x, axis=-1):
    return len(signal.find_peaks(x, axis=axis)[0])


def cross_zero_num(x, axis=-1):
    return len(np.where(np.diff(np.sign(x), axis=axis))[0])


def abs_max(x, axis=-1):
    return np.max(np.abs(x), axis=axis)


def mean(x, axis=-1):
    return np.mean(x, axis=axis)


def std(x, axis=-1):
    return np.std(x, axis=axis)


def skew(x, axis=-1):
    return stats.skew(x, axis=axis)


def kurtosis(x, axis=-1):
    return stats.kurtosis(x, axis=axis)


def power(x, axis=-1):
    return np.sum(np.square(x), axis=axis)


def diff_mean(x, axis=-1):
    return np.mean(np.diff(x, axis=axis), axis=axis)


def diff_std(x, axis=-1):
    return np.std(np.diff(x, axis=axis), axis=axis)



from .filter import bandpass_butter, notch
from .resample import resample_eeg
from .scale import scale_eeg
from .segment import segment_signal

__all__ = ['segment_signal', 'notch', 'bandpass_butter', 'scale_eeg', 'resample_eeg']



import os
import random
import shutil
import string


def random_string(length=10):
    """生成随机字符串"""
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


class DottableDict(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self


class RuntimeDir(os.PathLike):
    """管理运行时的临时目录"""

    RUNTIME_BASE_DIR = '.'

    def __init__(self, rm_on_exit=True, dir: str = None):
        # 如果 dir 为 None, 则会在 RUNTIME_BASE_DIR 下创建一个随机目录
        if dir is None:
            new_dir = os.path.join(self.RUNTIME_BASE_DIR, self._new_dir(rm_on_exit))
            while os.path.exists(new_dir):
                new_dir = os.path.join(self.RUNTIME_BASE_DIR, self._new_dir(rm_on_exit))
            os.makedirs(new_dir)
            self.dir = new_dir
        # 如果 dir 为 str, 则将改目录作为运行时目录
        else:
            if not os.path.exists(dir):
                os.makedirs(dir)
            self.dir = dir
        self.rm_on_exit = rm_on_exit
        self.file_handler = None

    @classmethod
    def set_base_dir(cls, base_dir):
        cls.RUNTIME_BASE_DIR = base_dir

    def _new_dir(self, rm_on_exit):
        new_dirname = random_string(8)
        return new_dirname

    def __str__(self) -> str:
        return self.dir

    __fspath__ = __str__

    __repr__ = __str__

    def write_file(self, file_name, content, *args, **wargs):
        """写入文件"""
        with open(os.path.join(self.dir, file_name), *args, **wargs) as self.file_handler:
            self.file_handler.write(content)

    def open_file(self, file_name, *args, **wargs):
        """打开文件"""
        self.file_handler = open(os.path.join(self.dir, file_name), *args, **wargs)
        return self.file_handler

    def delete(self):
        if self.file_handler is not None and not self.file_handler.closed:
            self.file_handler.close()
        if os.path.exists(self.dir):
            shutil.rmtree(self.dir)

    def purge(self):
        if self.rm_on_exit:
            self.delete()

    def __del__(self):
        self.purge()



import numpy as np


def percentile_digitalize(x, axis=None):
    q1 = np.quantile(x, 0.25, axis=axis)
    q2 = np.quantile(x, 0.5, axis=axis)
    q3 = np.quantile(x, 0.75, axis=axis)
    x_digit = np.where(x < q1, 0, np.where(x < q2, 1, np.where(x < q3, 2, 3)))
    return x_digit


def features_to_onehot(features, axis=0):
    features_digit = percentile_digitalize(features, axis=axis)
    features_onehot = np.eye(4)[features_digit]
    return features_onehot


def continuous_smoothing(one_hot, alpha=0.2, decay=0.25):
    classes = one_hot.shape[-1]
    max_idx = np.argmax(one_hot, axis=-1)[..., np.newaxis]
    max_idx = np.repeat(max_idx, classes, axis=-1)
    one_hot_idx = one_hot.copy()
    one_hot_idx[..., :] = np.arange(classes)
    delta = np.abs(one_hot_idx - max_idx)
    scale = np.power(decay, delta).sum(axis=-1, keepdims=True) - 1
    running_alpha = alpha / scale
    label = np.zeros((one_hot.shape), dtype=np.int32)
    label = one_hot * (1 - alpha) + (1 - one_hot) * running_alpha * np.power(decay, delta)
    return label



# -*- coding: utf-8 -*-
"""EEGNet Pytorch 实现
@Author       : Zuo YiPing
@Date         : 2022-12-05 16:57:43
@FilePath     : /zypml/zypml/module/backbone/eegnet.py
@LastEditTime : 2022-12-05 17:19:30
@Description  : https://github.com/bkozyrskiy/AML_EEG_challenge/blob/master/src/model_torch.py
"""

import math

import torch
from torch import nn
from torch.nn import functional as F

# from zyplib.torch.modules.conv import DepthWiseSplitConv2D


def convtransp_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1):
    """
    Utility function for computing output of transposed convolutions
    takes a tuple of (h,w) and returns a tuple of (h,w)
    """

    if type(h_w) is not tuple:
        h_w = (h_w, h_w)

    if type(kernel_size) is not tuple:
        kernel_size = (kernel_size, kernel_size)

    if type(stride) is not tuple:
        stride = (stride, stride)

    if type(pad) is not tuple:
        pad = (pad, pad)

    if type(dilation) is not tuple:
        dilation = (dilation, dilation)

    h = math.floor(
        (h_w[0] + 2 * pad[0] - dilation[0] * (kernel_size[0] - 1) - 1) / stride[0] + 1
    )
    w = math.floor(
        (h_w[1] + 2 * pad[1] - dilation[1] * (kernel_size[1] - 1) - 1) / stride[1] + 1
    )
    return h, w


def get_model_params(model):
    params_dict = {}
    for name, param in model.named_parameters():
        if param.requires_grad:
            params_dict[name] = param.data
    return params_dict


class deepwise_separable_conv(nn.Module):
    def __init__(self, nin, nout, kernelSize):
        super(deepwise_separable_conv, self).__init__()
        self.kernelSize = kernelSize
        self.time_padding = int(kernelSize // 2)
        self.depthwise = nn.Conv2d(
            in_channels=nin,
            out_channels=nin,
            kernel_size=(1, kernelSize),
            padding=(0, self.time_padding),
            groups=nin,
            bias=False,
        )
        self.pointwise = nn.Conv2d(
            in_channels=nin, out_channels=nout, kernel_size=1, groups=1, bias=False
        )

    def forward(self, input):
        dw = self.depthwise(input)
        pw = self.pointwise(dw)
        return pw

    def get_output_size(self, h_w):
        return convtransp_output_shape(
            h_w,
            kernel_size=(1, self.kernelSize),
            stride=1,
            pad=(0, self.time_padding),
            dilation=1,
        )


class EEGNet(nn.Module):
    """Data shape = (trials, kernels, channels, samples), which for the
    input layer, will be (trials, 1, channels, samples)."""

    # TODO resolve problems with avg padding when the end of the epoch lost
    # TODO possible solution via padding or AdaptiveAvgPool2d
    def __init__(
        self,
        nb_classes,
        Chans=64,
        Samples=128,
        dropoutRates=(0.25, 0.25),
        kernLength1=64,
        kernLength2=16,
        F1=4,
        D=2,
        F2=8,
        norm_rate=0.25,
    ):
        super(EEGNet, self).__init__()
        self.Chans = Chans
        self.Samples = Samples
        self.output_sizes = {}
        # block1
        time_padding = int((kernLength1 // 2))
        self.conv1 = nn.Conv2d(
            in_channels=1,
            out_channels=F1,
            kernel_size=(1, kernLength1),
            padding=(0, time_padding),
            stride=1,
            bias=False,
        )
        self.output_sizes['conv1'] = convtransp_output_shape(
            (Chans, Samples),
            kernel_size=(1, kernLength1),
            stride=1,
            pad=(0, time_padding),
        )
        self.batchnorm1 = nn.BatchNorm2d(num_features=F1, affine=True)
        self.depthwise1 = nn.Conv2d(
            in_channels=F1,
            out_channels=F1 * D,
            kernel_size=(Chans, 1),
            groups=F1,
            padding=0,
            bias=False,
        )
        self.output_sizes['depthwise1'] = convtransp_output_shape(
            self.output_sizes['conv1'], kernel_size=(Chans, 1), stride=1, pad=0
        )
        self.batchnorm2 = nn.BatchNorm2d(num_features=F1 * D, affine=True)
        self.activation_block1 = nn.ELU()
        # self.avg_pool_block1 = nn.AvgPool2d((1,poolKern1))
        # self.output_sizes['avg_pool_block1'] = convtransp_output_shape(self.output_sizes['depthwise1'], kernel_size=(1, poolKern1),
        #                                                           stride=(1,poolKern1), pad=0)
        self.avg_pool_block1 = nn.AdaptiveAvgPool2d(
            (1, int(self.output_sizes['depthwise1'][1] / 4))
        )
        self.output_sizes['avg_pool_block1'] = (
            1,
            int(self.output_sizes['depthwise1'][1] / 4),
        )
        self.dropout_block1 = nn.Dropout(p=dropoutRates[0])

        # block2
        self.separable_block2 = deepwise_separable_conv(
            nin=F1 * D, nout=F2, kernelSize=kernLength2
        )
        self.output_sizes['separable_block2'] = self.separable_block2.get_output_size(
            self.output_sizes['avg_pool_block1']
        )
        self.activation_block2 = nn.ELU()
        # self.avg_pool_block2 = nn.AvgPool2d((1,poolKern2))
        # self.output_sizes['avg_pool_block2'] = convtransp_output_shape(self.output_sizes['separable_block2'],
        #                                                                kernel_size=(1, poolKern2),
        #                                                                stride=(1, poolKern2), pad=0)
        self.avg_pool_block2 = nn.AdaptiveAvgPool2d(
            (1, int(self.output_sizes['separable_block2'][1] / 4))
        )
        self.output_sizes['avg_pool_block2'] = (
            1,
            int(self.output_sizes['separable_block2'][1] / 4),
        )

        self.dropout_block2 = nn.Dropout(dropoutRates[1])

        self.flatten = nn.Flatten()
        n_size = self.get_features_dim(Chans, Samples)
        self.dense = nn.Linear(n_size, nb_classes)

    def get_features_dim(self, Chans, Samples):
        bs = 1
        x = torch.rand((bs, 1, Chans, Samples))
        output_feat, out_dims = self.forward_features(x)
        n_size = output_feat.data.view(bs, -1).size(1)
        return n_size

    def forward_features(self, input):
        out_dims = {}
        block1 = self.conv1(input)
        out_dims['conv1'] = block1.size()
        block1 = self.batchnorm1(block1)
        block1 = self.depthwise1(block1)
        out_dims['depthwise1'] = block1.size()
        block1 = self.batchnorm2(block1)
        block1 = self.activation_block1(block1)
        block1 = self.avg_pool_block1(block1)
        out_dims['avg_pool_block1'] = block1.size()
        block1 = self.dropout_block1(block1)

        block2 = self.separable_block2(block1)
        out_dims['separable_block2'] = block1.size()
        block2 = self.activation_block2(block2)
        block2 = self.avg_pool_block2(block2)
        out_dims['avg_pool_block2'] = block1.size()
        block2 = self.dropout_block2(block2)
        return block2, out_dims

    def forward(self, input):
        features, _ = self.forward_features(input)
        flatten_feats = self.flatten(features)
        out = self.dense(flatten_feats)
        return out

    def weights_init(self):
        def weights_init(m):
            if isinstance(m, nn.Conv2d):
                torch.nn.init.xavier_uniform(m.weight.data)



import torch
from torch import nn


class SEBlock(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super(SEBlock, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_channels, in_channels // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(in_channels // reduction, in_channels, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x):
        b, c, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1)
        return x * y.expand_as(x)


class SEBlock2D(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super(SEBlock2D, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_channels, in_channels // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(in_channels // reduction, in_channels, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y.expand_as(x)



import random
from typing import Any, List, Union

import numpy as np


def random_split(
    data: Union[List[Any], np.ndarray],
    proportion: Union[List[float], List[int]],
    seed: int = None,
):
    """
    随机切分数据集; 类似 torch 的 random_split，不过返回的是切分的结果而非 Subset

    示例
    ----------
    ```py
    # first_part 占 70%，second_part 占 30%
    first_part, second_part = random_split(data, proportion=[0.7, 0.3])

    # 或根据具体数量进行划分，例如：
    first_part, second_part = random_split(data, proportion=[7, 3])
    ```

    Parameters
    ----------
    - `data` : `List[Any] | np.ndarray`
        - 需要切分的数据; 一个可以索引的对象
    - `proportion` : `List[float] | List[int]`
        - 切分的比例或每部分的数量:
          - 如果是 `float[]`，需要满足 `sum(proportion) == 1`；
          - 如果是 `int[]`，需要满足 `sum(proportion) == len(data)`。
    - `seed` : `int`, optional
        - 随机种子

    Returns
    -------
    - 切分后的多个部分: `List`
        - 按照比例或数量切分后的数据集部分。
    """

    if seed is not None:
        random.seed(seed)

    data_len = len(data)

    # 判断 proportion 是比例还是数量
    if all(isinstance(p, int) for p in proportion):
        # 如果是 int，确保数量之和等于数据长度
        if sum(proportion) != data_len:
            raise ValueError('整数形式的 proportion 的和必须等于数据总长度')
        sizes = proportion
    elif all(isinstance(p, float) for p in proportion):
        # 如果是 float，确保比例之和为 1
        if not np.isclose(sum(proportion), 1.0):
            raise ValueError('浮点数形式的 proportion 的和必须为 1')
        # 按比例计算每个部分的大小
        sizes = [int(p * data_len) for p in proportion]

        # 为了确保总大小与数据长度一致，调整最后一个部分的大小
        sizes[-1] = data_len - sum(sizes[:-1])
    else:
        raise TypeError('proportion 必须全为 float 或全为 int')

    # 随机打乱数据索引
    indices = list(range(data_len))
    random.shuffle(indices)

    # 根据 sizes 切分数据
    parts = []
    start_idx = 0
    for size in sizes:
        end_idx = start_idx + size
        part_indices = indices[start_idx:end_idx]
        part = (
            [data[i] for i in part_indices]
            if isinstance(data, list)
            else data[part_indices]
        )
        parts.append(part)
        start_idx = end_idx

    return parts



from abc import ABCMeta

from typing import Literal, Union

import torch
from torch import nn
from torch.nn import Module

XDim = Literal['1d', '2d', '3d', 1, 2, 3]
Conv = Union[nn.Conv1d, nn.Conv2d, nn.Conv3d]
BatchNorm = Union[nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d]
MaxPool = Union[nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d]
AvgPool = Union[nn.AvgPool1d, nn.AvgPool2d, nn.AvgPool3d]
AdaptiveAvgPool = Union[nn.AdaptiveAvgPool1d, nn.AdaptiveAvgPool2d, nn.AdaptiveAvgPool3d]
AdaptiveMaxPool = Union[nn.AdaptiveMaxPool1d, nn.AdaptiveMaxPool2d, nn.AdaptiveMaxPool3d]
ConvTranspose = Union[nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d]
Dropout = Union[nn.Dropout, nn.Dropout2d, nn.Dropout3d]


def conv(xdim: XDim = 1) -> Conv:
    if xdim == '1d':
        return nn.Conv1d
    elif xdim == '2d':
        return nn.Conv2d
    elif xdim == '3d':
        return nn.Conv3d
    elif xdim == 1:
        return nn.Conv1d
    elif xdim == 2:
        return nn.Conv2d
    elif xdim == 3:
        return nn.Conv3d
    else:
        raise ValueError(f'Invalid parameter xdim: {xdim}; 无法转换为对应的 Conv 类')


def batchnorm(xdim: XDim = 1) -> nn.Module:
    if xdim == '1d':
        return nn.BatchNorm1d
    elif xdim == '2d':
        return nn.BatchNorm2d
    elif xdim == '3d':
        return nn.BatchNorm3d
    elif xdim == 1:
        return nn.BatchNorm1d
    elif xdim == 2:
        return nn.BatchNorm2d
    elif xdim == 3:
        return nn.BatchNorm3d
    else:
        raise ValueError(f'Invalid parameter xdim: {xdim}; 无法转换为对应的 BatchNorm 类')


def maxpool(xdim: XDim = 1) -> MaxPool:
    if xdim == '1d':
        return nn.MaxPool1d
    elif xdim == '2d':
        return nn.MaxPool2d
    elif xdim == '3d':
        return nn.MaxPool3d
    elif xdim == 1:
        return nn.MaxPool1d
    elif xdim == 2:
        return nn.MaxPool2d
    elif xdim == 3:
        return nn.MaxPool3d
    else:
        raise ValueError(f'Invalid parameter xdim: {xdim}; 无法转换为对应的 MaxPool 类')


def avgpool(xdim: XDim = 1) -> AvgPool:
    if xdim == '1d':
        return nn.AvgPool1d
    elif xdim == '2d':
        return nn.AvgPool2d
    elif xdim == '3d':
        return nn.AvgPool3d
    elif xdim == 1:
        return nn.AvgPool1d
    elif xdim == 2:
        return nn.AvgPool2d
    elif xdim == 3:
        return nn.AvgPool3d
    else:
        raise ValueError(f'Invalid parameter xdim: {xdim}; 无法转换为对应的 AvgPool 类')


def adaptiveavgpool(xdim: XDim = 1) -> AdaptiveAvgPool:
    if xdim == '1d':
        return nn.AdaptiveAvgPool1d
    elif xdim == '2d':
        return nn.AdaptiveAvgPool2d
    elif xdim == '3d':
        return nn.AdaptiveAvgPool3d
    elif xdim == 1:
        return nn.AdaptiveAvgPool1d
    elif xdim == 2:
        return nn.AdaptiveAvgPool2d
    elif xdim == 3:
        return nn.AdaptiveAvgPool3d
    else:
        raise ValueError(
            f'Invalid parameter xdim: {xdim}; 无法转换为对应的 AdaptiveAvgPool 类'
        )


def adaptivemaxpool(xdim: XDim = 1) -> AdaptiveMaxPool:
    if xdim == '1d':
        return nn.AdaptiveMaxPool1d
    elif xdim == '2d':
        return nn.AdaptiveMaxPool2d
    elif xdim == '3d':
        return nn.AdaptiveMaxPool3d
    elif xdim == 1:
        return nn.AdaptiveMaxPool1d
    elif xdim == 2:
        return nn.AdaptiveMaxPool2d
    elif xdim == 3:
        return nn.AdaptiveMaxPool3d
    else:
        raise ValueError(
            f'Invalid parameter xdim: {xdim}; 无法转换为对应的 AdaptiveMaxPool 类'
        )


def convtranspose(xdim: XDim = 1) -> ConvTranspose:
    if xdim == '1d':
        return nn.ConvTranspose1d
    elif xdim == '2d':
        return nn.ConvTranspose2d
    elif xdim == '3d':
        return nn.ConvTranspose3d
    elif xdim == 1:
        return nn.ConvTranspose1d
    elif xdim == 2:
        return nn.ConvTranspose2d
    elif xdim == 3:
        return nn.ConvTranspose3d
    else:
        raise ValueError(
            f'Invalid parameter xdim: {xdim}; 无法转换为对应的 ConvTranspose 类'
        )


def dropout(xdim: XDim = 1) -> Dropout:
    if xdim == '1d':
        return nn.Dropout
    elif xdim == '2d':
        return nn.Dropout2d
    elif xdim == '3d':
        return nn.Dropout3d
    elif xdim == 1:
        return nn.Dropout
    elif xdim == 2:
        return nn.Dropout2d
    elif xdim == 3:
        return nn.Dropout3d
    else:
        raise ValueError(f'Invalid parameter xdim: {xdim}; 无法转换为对应的 Dropout 类')


class BaseModule(Module):
    __metaclass__ = ABCMeta

    def __hash__(self) -> int:
        return hash(str(self))

    def init_weights(self, weight_init_func=None):
        if weight_init_func is None:
            self.apply(self.default_init_weights)
        else:
            self.apply(weight_init_func)

    def default_init_weights(self, m):
        if isinstance(m, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
            nn.init.kaiming_uniform_(m.weight.data, nonlinearity='leaky_relu')
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
        elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)):
            nn.init.constant_(m.weight.data, 1)
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.Linear):
            nn.init.kaiming_uniform_(m.weight.data)
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)



import lightning as L
import torch
from torch import nn


class BasicClassifier(L.LightningModule):
    def __init__(
        self,
        model: nn.Module,
        optimizer: torch.optim.Optimizer,
        loss_fn: nn.Module = None,
        metrics: dict = None,
    ):
        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.loss_fn = loss_fn or nn.CrossEntropyLoss()
        self.metrics = metrics or {}

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.loss_fn(y_hat, y)
        self.log('train_loss', loss)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.loss_fn(y_hat, y)
        self.log('val_loss', loss)
        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.loss_fn(y_hat, y)
        self.log('test_loss', loss)
        return loss

    def configure_optimizers(self):
        return self.optimizer

    def on_train_epoch_end(self):
        for metric_name, metric in self.metrics.items():
            self.log(f'Train/{metric_name}', metric.compute())
            metric.reset()

    def on_validation_epoch_end(self):
        for metric_name, metric in self.metrics.items():
            self.log(f'Val/{metric_name}', metric.compute())
            metric.reset()

    def on_test_epoch_end(self):
        for metric_name, metric in self.metrics.items():
            self.log(f'Test/{metric_name}', metric.compute())
            metric.reset()



"""
本模块主要提供了一些数据集导入的 pipeline 相关的函数，是对脑电信号数据导入、预处理流程的高级封装

所有函数均满足:

- 输入: 一些参数，以及一些回调函数
- 输出: 返回数据集 (X, y)

"""
from functools import reduce
from typing import Callable, Union

import numpy as np

__all__ = ['basic_pipeline']



def basic_pipeline(
    files: list[str],
    fn_read: Callable[[str], np.ndarray],
    fn_segment: Callable[[np.ndarray], np.ndarray],
    label: Union[int, Callable[[str], int]],
    fn_before_segment: Callable[[np.ndarray], np.ndarray] = None,
    fn_after_segment: Callable[[np.ndarray], np.ndarray] = None,
):
    """从文件中读取数据，并自动切片、制作用于训练的数据集

    本 Pipeline 适用于:
    - 从文件中读取完整的脑电
    - 需要对完整脑电进行切分
    - 无特殊标签，只需根据文件进行标记即可
    - 输出: 数据集 (X, y)
        - X: 数据，shape = (N, C, T)
        - y: 标签，shape = (N,)

    Pipeline 流程:
    ----------

    1. 从文件中读取完整的数据 X
    2. 对数据 X 进行预处理
    3. 对数据 X 进行切片
    4. 对数据 X 进行后处理
    5. 制作标签 y
    6. 返回数据集 (X, y)

    Parameters
    ----------
    - `files` : `list[str]`
        - 文件路径列表
    - `fn_read` : `(file_path: str) -> np.ndarray`
        - 文件读取函数
    - `fn_segment` : `(signal: np.ndarray) -> np.ndarray`
        - 数据切片函数
        - 输入一个 `[C, T]` 的信号
        - 输出一个 `[N, C, T]` 的信号
    - `label` : `int | (file_path: str) -> int`
        - 标签，如果为 int，则所有数据标签相同；如果为函数，则根据文件名生成标签
    - `fn_before_segment` : `(signal: [C, T]) -> [C, T]`, optional
        - 数据切片前的处理函数
    - `fn_after_segment` : `(signal: [N, C, T]) -> [N, C, T]`, optional
        - 数据切片后的处理函数

    Returns
    ----------
    - `X` : `np.ndarray`, shape = (N, C, T)
        - 数据
    - `y` : `np.ndarray`, shape = (N,)
        - 标签
    """
    if not files:
        raise ValueError('The files list is empty. Please provide at least one file.')

    fn_before_segment = fn_before_segment or (lambda x: x)
    fn_after_segment = fn_after_segment or (lambda x: x)
    label = label if callable(label) else (lambda _: label)

    # 处理单个文件
    def load_single_file(fpath: str):
        signal = fn_read(fpath)  # 读取数据
        signal = fn_before_segment(signal)  # 数据预处理
        segments = fn_segment(signal)  # 数据切片
        segments = fn_after_segment(segments)
        N = len(segments)
        label_ = label(fpath)  # 制作标签
        return segments, [label_] * N

    # 批量处理所有文件
    results = map(load_single_file, files)

    # 合并数据
    def merge_data(Xy: tuple[list[np.ndarray], list[int]], pair: tuple[np.ndarray, int]):
        Xy[0].append(pair[0])
        Xy[1].extend(pair[1])
        return Xy

    Xy = reduce(merge_data, results, ([], []))

    X = np.concatenate(Xy[0], axis=0)
    y = np.array(Xy[1], dtype=int)
    return X, y



import time
from datetime import datetime, timedelta, timezone
from timeit import default_timer as timer


def cur_time(pat: str ='%m/%d %H:%M:%S'):
    ts = time.time()
    td = timedelta(hours=8)  # 我们在东八区
    tz = timezone(td)
    dt = datetime.fromtimestamp(ts, tz)
    dt = dt.strftime(pat)
    return dt


class TicToc:
    def __init__(self):
        self._tic = 0
        self._toc = 0
        self._elapsed = 0

    def tic(self):
        self._tic = timer()

    def toc(self):
        self._toc = timer()
        self._elapsed = self._toc - self._tic

    @property
    def elapsed(self):
        delta = timedelta(seconds=self._elapsed)
        return str(delta)[:-3]



from typing import Literal

COLOR_MAP = {
    'red': '\033[91m',
    'green': '\033[92m',
    'yellow': '\033[93m',
    'blue': '\033[94m',
    'magenta': '\033[95m',
    'cyan': '\033[96m',
    'white': '\033[97m',
    'reset': '\033[0m',
}


COLOR_TYPE = Literal['red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white']


def colored(text: str, color: COLOR_TYPE):
    """为文本添加颜色


    Parameters
    ----------
    - `text` : `str`
        - 原始文本
    - `color` : `COLOR_TYPE`
        - 颜色类型
        - `"red"` | `"green"` | `"yellow"` | `"blue"` | `"magenta"` | `"cyan"` | `"white"`
    """
    return f"{COLOR_MAP[color]}{text}{COLOR_MAP['reset']}"


def print_colored(color: COLOR_TYPE, *args, **kwargs):
    """打印带颜色的文本

    Parameters
    ----------
    - `color` : `COLOR_TYPE`
        - 颜色类型
        - `"red"` | `"green"` | `"yellow"` | `"blue"` | `"magenta"` | `"cyan"` | `"white"`
    """

    total_text = ' '.join([str(arg) for arg in args])
    print(colored(total_text, color), **kwargs)


def debug(*args, **kwargs):
    """打印调试信息（白色）"""
    print_colored('white', *args, **kwargs)


def info(*args, **kwargs):
    """打印信息（青色）"""
    print_colored('cyan', *args, **kwargs)


def warn(*args, **kwargs):
    """打印警告信息（黄色）"""
    print_colored('yellow', *args, **kwargs)


def error(*args, **kwargs):
    """打印错误信息（红色）"""
    print_colored('red', *args, **kwargs)



import numpy as np
from plotly import express as px
from plotly import graph_objects as go

from ..utils.ensure import ensure_2dims, ensure_npy


def plotly_waves(
    signals,
    srate=None,
    channels_name=None,
    fixed_y=False,
    colorscheme='Prism',
    colors: list = None,
    height=None,
    width=None,
    show=True,
):
    """使用 Plotly 可视化多通道信号


    Parameters
    ----------
    - `signals` : _type_
        - _description_
    - `srate` : _type_, optional
        - _description_, by default None
    - `channels_name` : _type_, optional
        - _description_, by default None
    - `fixed_y` : bool, optional
        - _description_, by default False
    - `colorscheme` : str, optional
        - _description_, by default 'Prism'
    - `colors` : list, optional
        - _description_, by default None
    - `height` : _type_, optional
        - _description_, by default None
    - `width` : _type_, optional
        - _description_, by default None
    - `show` : bool, optional
        - _description_, by default True

    Returns
    ----------
    - `_type_`
        - _description_
    """
    signals = ensure_npy(signals)
    signals = ensure_2dims(signals)

    ch_cnt, time_cnt = signals.shape
    if channels_name is None:
        channels_name = [f'Ch{i}' for i in range(ch_cnt)]

    x = np.arange(time_cnt)
    if srate:
        x = x / srate
    np.linspace(0, 1, ch_cnt + 1)
    y_part = np.linspace(0, 1, ch_cnt + 1)
    # y_part = list(reversed(np.linspace(0, 1, ch_cnt + 1)))

    if colors is None:
        colors = getattr(px.colors.qualitative, colorscheme)
    elif isinstance(colors, str):
        colors[colors] * ch_cnt

    fig = go.Figure()

    for ch in range(ch_cnt):
        y = signals[ch, :]
        name = channels_name[ch]
        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                name=name,
                mode='lines',
                yaxis=f'y{ch + 1}',
                marker=dict(color=colors[ch]),
            )
        )
        yaxis_key = f'yaxis{ch + 1}'  # if ch > 0 else 'yaxis'
        yaxis = dict(
            automargin=True,
            showline=True,
            side='left',
            domain=[y_part[ch], y_part[ch + 1]],
            # domain=[y_part[ch + 1], y_part[ch]],
            fixedrange=fixed_y,
            title=name,
            mirror=True,
            type='linear',
            linecolor=colors[ch],
            # showticklabels=False,
            # nticks=3,
            tickfont={'color': colors[ch], 'size': 10},
            tickmode='auto',
            titlefont={'color': colors[ch]},
        )
        fig.update_layout(**{yaxis_key: yaxis})

    fig.update_traces(hoverinfo='x+name+y', showlegend=False)

    fig.update_xaxes(
        autorange=True,
        rangeslider=dict(autorange=True, visible=True),
        showticklabels=True,
        tickfont={'size': 12},
    )

    fig.update_layout(
        autosize=True,
        height=height,
        width=width,
        hovermode='x',
        template='plotly_white',
        legend=dict(traceorder='reversed'),
        margin=dict(t=0, b=0),
    )
    if show:
        fig.show()
    return fig



"""
同构建模型训练数据集相关的模块
"""

from .make_xy_pipeline import basic_pipeline
from .split import random_split
from .torch import BasicXyDataset as TorchBasicXyDataset
from .torch import make_dataloader_from_xy

__all__ = [
    'basic_pipeline',
    'random_split',
    'TorchBasicXyDataset',
    'make_dataloader_from_xy',
]



import json
import os
from dataclasses import dataclass

BASE_DIR = os.path.expanduser('~/.zyplib')
CONFIG_PATH = os.path.join(BASE_DIR, 'config.json')


@dataclass
class Config:
    DISK_CACHE_DIR: str = '~/.zyplib/cache'


def _write_config(config: Config):
    if not os.path.exists(BASE_DIR):
        os.makedirs(BASE_DIR)

    config_path = os.path.expanduser(CONFIG_PATH)
    with open(config_path, 'w', encoding='utf-8') as file:
        json.dump(config.__dict__, file, indent=4)


def _load_config() -> Config:
    # 获取配置文件路径
    config_path = os.path.expanduser(CONFIG_PATH)

    # 如果配置文件存在，则加载配置文件
    if os.path.exists(config_path):
        with open(config_path, 'r', encoding='utf-8') as file:
            config_dict = json.load(file)

        # 覆盖默认配置
        return Config(**config_dict)
    else:
        # 如果配置文件不存在，返回默认配置
        config = Config()
        _write_config(config)
        return config


# 实例化配置，在模块导入时加载
config = _load_config()



from typing import List, Union

import numpy as np
import mne
from .ensure import ensure_npy
from zyplib.annotation.event import Event


def to_mne_raw(data, srate: float, channels: List[str], events: List[Event] = None):
    info = mne.create_info(channels, srate, ch_types='eeg', verbose=False)
    raw = mne.io.RawArray(data, info)
    if events is not None:
        onset = np.array([event.onset for event in events])
        duration = np.array([event.duration for event in events])
        description = np.array([event.name for event in events])
        raw.set_annotations(
            mne.Annotations(onset=onset, duration=duration, description=description)
        )
    return raw



from typing import Sequence, Union

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
from torch.utils.data import random_split as torch_random_split

from zyplib.utils.ensure import ensure_tensor


class BasicXyDataset(Dataset):
    def __init__(
        self,
        data: Union[np.ndarray, torch.Tensor],
        labels: Union[np.ndarray, torch.Tensor],
    ):
        self.data = ensure_tensor(data)
        self.labels = ensure_tensor(labels)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]


def make_dataloader_from_xy(
    xy_train: tuple[np.ndarray, np.ndarray],
    xy_test: tuple[np.ndarray, np.ndarray] = None,
    split_val_from_train: Sequence[Union[int, float]] = None,
    batch_size: int = 64,
    shuffle: bool = True,
    num_workers: int = 0,
    pin_memory: bool = False,
    **kwargs,
):
    """从训练和测试数据创建 DataLoader(s)。
    
    参数
    ----------
    - `xy_train` : `tuple[np.ndarray, np.ndarray]`
        - 包含训练数据和标签的元组。
    - `xy_test` : `tuple[np.ndarray, np.ndarray]`, 可选
        - 包含测试数据和标签的元组，默认为 None。
    - `split_val_from_train` : `tuple[float, float] | tuple[int, int]`, 可选
        - 指定如何将训练数据拆分为训练集和验证集，默认为 None。
        - 如果为 None，则不进行拆分。
    - `batch_size` : `int`, 可选
        - 每个批次的样本数量，默认为 64。
    - `shuffle` : `bool`, 可选
        - 是否在每个 epoch 时打乱数据，默认为 True。
    - `num_workers` : `int`, 可选
        - 用于数据加载的子进程数量，默认为 0。
    - `pin_memory` : `bool`, 可选
        - 如果为 True，数据加载器将把张量复制到 CUDA 固定内存中，默认为 False。
    - `**kwargs` : `dict`, 可选
        - DataLoader 的其他参数，默认为 None。
    
    返回
    ----------
    - `tuple[DataLoader]`
        - 训练、验证和测试数据集的 DataLoader 元组。
    """
    train_dataset = BasicXyDataset(*xy_train)

    if xy_test is not None:
        test_dataset = BasicXyDataset(*xy_test)
    else:
        test_dataset = None

    # 划分验证集
    if split_val_from_train:
        train_dataset, val_dataset = torch_random_split(
            train_dataset, split_val_from_train
        )
    else:
        val_dataset = None

    def make_dataloader(dataset: Dataset):
        if dataset is None:
            return None
        return DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory,
            **kwargs,
        )

    loaders = map(make_dataloader, [train_dataset, val_dataset, test_dataset])
    loaders = filter(lambda x: x is not None, loaders)
    return tuple(loaders)
